/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 * Author: Damien Fournier
 */

package org.ow2.frascati.tinfi.emf;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.Resource.Factory.Registry;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.stp.sca.BaseReference;
import org.eclipse.stp.sca.BaseService;
import org.eclipse.stp.sca.Component;
import org.eclipse.stp.sca.ComponentReference;
import org.eclipse.stp.sca.ComponentService;
import org.eclipse.stp.sca.Composite;
import org.eclipse.stp.sca.DocumentRoot;
import org.eclipse.stp.sca.Implementation;
import org.eclipse.stp.sca.Interface;
import org.eclipse.stp.sca.JavaImplementation;
import org.eclipse.stp.sca.JavaInterface;
import org.eclipse.stp.sca.Multiplicity;
import org.eclipse.stp.sca.PropertyValue;
import org.eclipse.stp.sca.Reference;
import org.eclipse.stp.sca.SCAImplementation;
import org.eclipse.stp.sca.ScaPackage;
import org.eclipse.stp.sca.Service;
import org.eclipse.stp.sca.Wire;
import org.eclipse.stp.sca.util.ScaResourceFactoryImpl;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.api.type.TypeFactory;
import org.objectweb.fractal.julia.type.BasicComponentType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.ClassGeneratorItf;
import org.objectweb.fractal.juliac.FCSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.JuliacRuntimeException;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.AttributeDesc;
import org.objectweb.fractal.juliac.desc.BindingDesc;
import org.objectweb.fractal.juliac.desc.BindingType;
import org.objectweb.fractal.juliac.desc.ComponentDescFactoryClassGenerator;
import org.ow2.frascati.model.osgi.OsgiImplementation;
import org.ow2.frascati.model.osgi.OsgiPackage;
import org.ow2.frascati.tinfi.TinfiDomain;

/**
 * EMF parser for SCA Assembly Language descriptors.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Damien Fournier <Damien.Fournier@inria.fr>
 * @since 1.1
 */
public class EMFParserSupportImpl {

    /**
     * Initialize the service.
     * 
     * @param jc  the instance of Juliac used by this service
     */
    public void init( Juliac jc ) {
        this.jc = jc;
    }
    private Juliac jc;
    private int compidx = 0;
    

    /**
     * Parse the specified ADL descriptor and return a model of it.
     * 
     * @param adl  the fully-qualified name of the ADL descriptor
     * @return     the model of the specified ADL
     */
    public <T> T parse( String adl ) throws IOException {
        
        JuliacConfig jconf = jc.getJuliacConfig();
        ClassLoader loader = jconf.getClassLoader();
        
        String adlFileName = adl.replace('.','/')+".composite";
        URL url = loader.getResource(adlFileName);
        if( url == null ) {
            String msg = "Can not load "+adlFileName;
            throw new IOException(msg);
        }
        
        URI uri;
        try {
            uri = url.toURI();
        }
        catch (URISyntaxException e) {
            IOException ioe = new IOException();
            ioe.initCause(e);
            throw ioe;
        }        
        ResourceSet resourceSet = new ResourceSetImpl();

        /*
         * Register the appropriate resource factory to handle all file
         * extensions.
         */
        Registry resFactReg = resourceSet.getResourceFactoryRegistry();
        Map<String,Object> map = resFactReg.getExtensionToFactoryMap();
        Object scaResFact = new ScaResourceFactoryImpl();
        map.put("*",scaResFact);

        /*
         * Register the EMF packages.
         */
        org.eclipse.emf.ecore.EPackage.Registry reg =
            resourceSet.getPackageRegistry();
        reg.put( ScaPackage.eNS_URI, ScaPackage.eINSTANCE );
        reg.put( OsgiPackage.eNS_URI, OsgiPackage.eINSTANCE );
                
        /*
         * Parse the composite file.
         */
        org.eclipse.emf.common.util.URI resourceUri =
            org.eclipse.emf.common.util.URI.createURI(uri.toString());
        Resource resource = resourceSet.getResource(resourceUri,true);

        /*
         * Get the root node of the composite model.
         */
        EList<EObject> list = resource.getContents();
        DocumentRoot root = (DocumentRoot) list.get(0);
        Composite composite = root.getComposite();

        @SuppressWarnings("unchecked")
        T ret = (T) composite;
        return ret;
    }

    /**
     * Generate the source code of a factory class for the specified ADL
     * descriptor.
     * 
     * @param adl         the fully-qualified name of the ADL descriptor
     * @param targetname  the fully-qualified name of the factory class
     */
    public void generate( String adl, String targetname ) throws IOException {
        
        /*
         * Generate the factory for the assembly.
         */
        ComponentDesc cdesc = toComponentDesc(adl);
        ClassGeneratorItf cg =
            new ComponentDescFactoryClassGenerator(targetname,cdesc);
        jc.generateSourceCode(cg);
        
        /*
         * Generate factories for each component included in the assembly.
         */
        generateIOE(cdesc);
    }
    
    
    // ----------------------------------------------------------------------
    // Methods for transforming Composite to ComponentDesc
    // ----------------------------------------------------------------------
    
    /**
     * Return the {@link ComponentDesc} corresponding to the specified ADL
     * descriptor.
     */
    public ComponentDesc toComponentDesc( String adl ) throws IOException {
        Composite composite = parse(adl);
        ComponentDesc cdesc = toComponentDesc(composite);
        return cdesc;
    }
    
    /**
     * Transform the specified {@link Composite} to {@link ComponentDesc}.
     */
    private ComponentDesc toComponentDesc( Composite composite )
    throws IOException {
        
        String name = composite.getName();
        EList<Component> subs = composite.getComponent();

        /*
         * Composite type.
         */
        String id = "C"+(compidx++);
        ComponentType ct = ecore2fractal(composite);
        String controllerDesc = TinfiDomain.SCA_COMPOSITE;

        FCSourceCodeGeneratorItf<?> fcscg =
            jc.getFCSourceCodeGenerator(ct,controllerDesc,null);
        
        ComponentDesc cdesc =
            new ComponentDesc(id,name,name,fcscg,ct,controllerDesc,null);

        /*
         * Subcomponents.
         */
        Map<String,ComponentDesc> subcdescs =
            new HashMap<String,ComponentDesc>();
        for (Component sub : subs) {
            ComponentDesc subcdesc = handleComponent(name,sub);
            cdesc.addSubComponent(subcdesc);
            subcdesc.addSuperComponent(cdesc);
            String subname = sub.getName();
            subcdescs.put(subname,subcdesc);
        }
        
        /*
         * Export bindings.
         * Promotion links between a service declared by a subcomponent and a
         * service declared by the composite.
         */
        EList<Service> services = composite.getService();
        for (Service service : services) {
            handleExport(cdesc,name,service,subcdescs);
        }
        
        /*
         * Import bindings.
         * Promotion links between a reference declared by a subcomponent and a
         * reference declared by the composite.
         */
        EList<Reference> references = composite.getReference();
        for (Reference reference : references) {
            handleImport(cdesc,name,reference,subcdescs);
        }
        
        /*
         * Internal bindings.
         * Internal links between two subcomponents. There are similar to wires
         * except that they are declared with the <code>target</code> parameter
         * of the <code>&lt;reference&gt;</code> tag.
         */
        for (Component sub : subs) {
            handleInternal(name,sub,subcdescs);
        }
                
        /*
         * Wires.
         * Internal links declared with the <code>&lt;wire&gt;</code> tag
         * between two subcomponents of the composite.
         */
        EList<Wire> wires = composite.getWire();
        for (Wire wire : wires) {
            handleWire(wire,subcdescs);
        }
        
        return cdesc;
    }

    private ComponentDesc handleComponent( String compositeName, Component sub )
    throws IOException {
        
        /*
         * Component name and type.
         */
        String name = sub.getName();
        ComponentType subct = ecore2fractal(sub);        
        Implementation implementation = sub.getImplementation();
        ComponentDesc cdesc = null;
        
        if( implementation instanceof JavaImplementation ) {

            /*
             * Java implementation.
             */
            JavaImplementation jimpl = (JavaImplementation) implementation;
            String controllerDesc = TinfiDomain.SCA_PRIMITIVE;
            String contentDesc = jimpl.getClass_();
            FCSourceCodeGeneratorItf<?> fcscg =
                jc.getFCSourceCodeGenerator(subct,controllerDesc,contentDesc);            
            
            String id = "C"+(compidx++);
            cdesc =
                new ComponentDesc(
                    id,name,name,fcscg,subct,controllerDesc,contentDesc);
        }
        else if( implementation instanceof SCAImplementation ) {
            
            /*
             * Composite implementation.
             */
            
            SCAImplementation scaimpl = (SCAImplementation) implementation;
            QName qname = scaimpl.getName();

            String namespaceURI = qname.getNamespaceURI();
            String localpart = qname.getLocalPart();
            if( ! localpart.endsWith(".composite") ) {
                String msg =
                    "Unexpected name (local part should end with .composite "+
                    qname+" in implementation.composite for subcomponent "+
                    sub.getName()+" in composite "+compositeName;
                throw new IOException(msg);
            }
            String fname = localpart.substring(0,localpart.length()-10);  // Remove .composite
            String adl = namespaceURI.replace('/','.')+'.'+fname;
            
            cdesc = toComponentDesc(adl);
        }
        else if( implementation instanceof OsgiImplementation ) {

            /*
             * OSGi bundle implementation.
             */
            OsgiImplementation osgiimpl = (OsgiImplementation) implementation;
            String controllerDesc = "osgiPrimitive";
            String contentDesc = osgiimpl.getBundle();
            FCSourceCodeGeneratorItf<?> fcscg =
                jc.getFCSourceCodeGenerator(subct,controllerDesc,contentDesc);            
            
            String id = "C"+(compidx++);
            cdesc =
                new ComponentDesc(
                    id,name,name,fcscg,subct,controllerDesc,contentDesc);
        }
        else {
            String msg =
                "Unsupported implementation type "+
                implementation.getClass().getName()+ " for subcomponent "+
                sub.getName()+" in composite "+compositeName;
            throw new IOException(msg);
        }
        
        /*
         * Component properties.
         */
        EList<PropertyValue> propvalues = sub.getProperty();
        for (PropertyValue propvalue : propvalues) {
            handleProperty(cdesc,propvalue);                
        }

        return cdesc;
    }
    
    private void handleExport(
        ComponentDesc cdesc, String compositeName,
        Service service, Map<String,ComponentDesc> subcdescs )
    throws IOException {
        
        String serviceName = service.getName();
        String promote = service.getPromote();
        String[] promotes = promote.split("/");
        if( promotes.length != 2 ) {
            String msg =
                "Illegal service promotion "+promote+" in composite "+
                compositeName;
            throw new IOException(msg);
        }
        
        ComponentDesc target = subcdescs.get(promotes[0]);
        if( target == null ) {
            String msg =
                "No such target component "+promotes[0]+
                " in service promotion "+promote+" in composite "+compositeName;
            throw new IOException(msg);
        }
        
        createBinding(BindingType.EXPORT,cdesc,serviceName,target,promotes[1]);
    }
    
    private void handleImport(
        ComponentDesc cdesc, String compositeName,
        Reference reference, Map<String,ComponentDesc> subcdescs )
    throws IOException {
        
        String referenceName = reference.getName();
        String promote = reference.getPromote();
        String[] promotes = promote.split(" ");
        for (String pr : promotes) {
            String[] prs = pr.split("/");
            if( prs.length != 2 ) {
                String msg =
                    "Illegal reference promotion "+promote+" in composite "+
                    compositeName;
                throw new IOException(msg);
            }
            
            ComponentDesc src = subcdescs.get(prs[0]);
            if( src == null ) {
                String msg =
                    "No such source component "+prs[0]+
                    " in reference promotion "+promote+" in composite "+
                    compositeName;
                throw new IOException(msg);
            }
            
            createBinding(BindingType.IMPORT,src,prs[1],cdesc,referenceName);
        }
    }
        
    private void handleWire( Wire wire, Map<String,ComponentDesc> subcdescs )
    throws IOException {
        
        // Wire source
        String source = wire.getSource();
        String[] sources = source.split("/");
        if( sources.length != 2 ) {
            String msg = "Illegal source "+source+" in wire "+wire;
            throw new IOException(msg);
        }
        ComponentDesc clt = subcdescs.get(sources[0]);
        if( clt == null ) {
            String msg =
                "No such source component "+sources[0]+" in wire "+wire;
            throw new IOException(msg);
        }
        
        // Wire target
        String target = wire.getTarget();
        String[] targets = target.split("/");
        if( targets.length != 2 ) {
            String msg = "Illegal target "+target+" in wire "+wire;
            throw new IOException(msg);
        }
        ComponentDesc srv = subcdescs.get(targets[0]);
        if( srv == null ) {
            String msg =
                "No such source component "+targets[0]+" in wire "+wire;
            throw new IOException(msg);
        }
        
        // Create binding
        createBinding(BindingType.NORMAL,clt,sources[1],srv,targets[1]);
    }
    
    private void handleInternal(
        String compositeName, Component sub,
        Map<String,ComponentDesc> subcdescs )
    throws IOException {
        
        String name = sub.getName();
        ComponentDesc clt = subcdescs.get(name);
        
        EList<ComponentReference> references = sub.getReference();
        for (ComponentReference reference : references) {
            
            String target = reference.getTarget();
            if( target!=null && target.length()>0 ) {
                                
                // Target(s)
                String[] targets = target.split(" ");
                int i = 0;
                for (String targ : targets) {
                    
                    // Source
                    String cltItfName = reference.getName();
                    if( targets.length > 1 ) {
                        /*
                         * In case of cardinality 0..n or 1..n for the targets,
                         * add an auto-incremented suffix to the source
                         * interface name to obtain the unique identifier needed
                         * by Julia Binding Controller.
                         */
                        cltItfName += i;
                    }
                    
                    String[] targs = targ.split("/");
                    if( targs.length != 2 ) {
                        String msg =
                            "Illegal reference "+target+" in composite "+
                            compositeName;
                        throw new IOException(msg);
                    }
                    ComponentDesc srv = subcdescs.get(targs[0]);
                    if( srv == null ) {
                        String msg =
                            "No such target component "+targs[0]+
                            " in reference "+target+" in composite "+
                            compositeName;
                        throw new IOException(msg);
                    }
                    
                    // Create binding
                    createBinding(
                        BindingType.NORMAL,clt,cltItfName,srv,targs[1]);
                    i++;
                }
            }
        }
    }
    
    private void createBinding(
        BindingType bindingType, ComponentDesc clt, String cltItfName,
        ComponentDesc srv, String srvItfName ) {
        
        BindingDesc bdesc =
            new BindingDesc(bindingType,clt,cltItfName,srv,srvItfName);
        clt.addBinding(cltItfName,bdesc);
    }
    
    private void handleProperty(
        ComponentDesc cdesc, PropertyValue propvalue ) {

        String name = propvalue.getName();
        String value = propvalue.getValue();
        AttributeDesc adesc = new AttributeDesc(cdesc,null,name,value);
        cdesc.addAttribute(name,adesc);
    }
    
    /**
     * Return the Fractal component type corresponding to the specified SCA
     * ecore composite.
     * 
     * @param composite  the SCA ecore composite
     * @return           the Fractal component type
     * @throws IOException  if the Fractal component type can not be created
     */
    private ComponentType ecore2fractal( Composite composite )
    throws IOException {
        
        EList<Service> services = composite.getService();
        EList<Reference> references = composite.getReference();
        InterfaceType[] its =
            new InterfaceType[services.size()+references.size()];
        int i = 0;
        
        for (Service service : services) {
            its[i] = ecore2fractal(service);
            i++;
        }
        for (Reference reference : references) {
            its[i] = ecore2fractal(reference);
            i++;
        }
        
        ComponentType ct = getComponentTypeIOE(its);
        return ct;
    }
    
    /**
     * Return the Fractal component type corresponding to the specified SCA
     * ecore composite.
     * 
     * @param component  the SCA ecore composite
     * @return           the Fractal component type
     * @throws IOException  if the Fractal component type can not be created
     */
    private ComponentType ecore2fractal( Component component )
    throws IOException {
        
        EList<ComponentService> services = component.getService();
        EList<ComponentReference> references = component.getReference();
        InterfaceType[] its =
            new InterfaceType[services.size()+references.size()];
        int i = 0;
        
        for (BaseService service : services) {
            its[i] = ecore2fractal(service);
            i++;
        }
        for (BaseReference reference : references) {
            its[i] = ecore2fractal(reference);
            i++;
        }
        
        ComponentType ct = getComponentTypeIOE(its);
        return ct;
    }
    
    /**
     * Return the Fractal interface type corresponding to the specified SCA
     * ecore service.
     * 
     * @param service  the SCA ecore service
     * @return         the Fractal interface type
     */
    private InterfaceType ecore2fractal( BaseService service ) {
        String name = service.getName();
        Interface itf = service.getInterface();
        InterfaceType it =
            ecore2fractal(name,itf,ONE_TO_ONE,TypeFactory.SERVER);
        return it;        
    }
    
    /**
     * Return the Fractal interface type corresponding to the specified SCA
     * ecore reference.
     * 
     * @param service  the SCA ecore reference
     * @return         the Fractal interface type
     */
    private InterfaceType ecore2fractal( BaseReference reference ) {
        String name = reference.getName();
        Interface itf = reference.getInterface();
        Multiplicity multiplicity = reference.getMultiplicity();
        String mult = multiplicity.getLiteral();
        InterfaceType it = ecore2fractal(name,itf,mult,TypeFactory.CLIENT);
        return it;        
    }
    
    /**
     * Return the component type corresponding to the specified interface types.
     * 
     * @param its  the interface types
     * @return     the component type
     * @throws IOException  if the component type can not be created
     */
    private ComponentType getComponentTypeIOE( InterfaceType[] its )
    throws IOException {
        try {
            ComponentType ct = new BasicComponentType(its);
            return ct;
        }
        catch (InstantiationException e) {
            IOException ioe = new IOException();
            ioe.initCause(e);
            throw ioe;
        }
    }
    
    /**
     * Return the Fractal interface type corresponding to the specified SCA
     * ecore interface definition.
     * 
     * @param name      the interface name
     * @param itf       the SCA ecore interface definition
     * @param mult      the interface multiplicity (0..1, 1..1, 0..n or 1..n)
     * @param isClient  <code>true</code> if the interface is required
     *                  <code>false</code> if the interface is provided
     * @return          the corresponding Fractal interface type
     * @throws JuliacRuntimeException
     *      if the interface multiplicity if not of the four following legal
     *      values: 0..1, 1..1, 0..n or 1..n 
     */
    private InterfaceType ecore2fractal(
        String name, Interface itf, String mult, boolean isClient )
    throws JuliacRuntimeException {

        if( ! cardinalities.containsKey(mult) ) {
            String msg = "Unsupported interface multiplicity: "+mult;
            throw new JuliacRuntimeException(msg);
        }
        
        JavaInterface jitf = (JavaInterface) itf;
        String signature = jitf.getInterface();
        boolean isOptional = contingencies.get(mult);
        boolean isCollection = cardinalities.get(mult);
        
        InterfaceType it =
            new BasicInterfaceType(
                name,signature,isClient,isOptional,isCollection);
        return it;
    }
    
    final private static String ZERO_TO_ONE = "0..1";
    final private static String ONE_TO_ONE = "1..1";
    final private static String ZERO_TO_N = "0..n";
    final private static String ONE_TO_N = "1..n";
    
    final private static Map<String,Boolean> contingencies =
        new HashMap<String,Boolean>() {
            private static final long serialVersionUID = 1306190174200744738L;
        {
            put(ZERO_TO_ONE,TypeFactory.OPTIONAL);
            put(ONE_TO_ONE,TypeFactory.MANDATORY);
            put(ZERO_TO_N,TypeFactory.OPTIONAL);
            put(ONE_TO_N,TypeFactory.MANDATORY);
        }};

    final private static Map<String,Boolean> cardinalities =
        new HashMap<String,Boolean>() {
            private static final long serialVersionUID = 1306190174200744738L;
        {
            put(ZERO_TO_ONE,TypeFactory.SINGLE);
            put(ONE_TO_ONE,TypeFactory.SINGLE);
            put(ZERO_TO_N,TypeFactory.COLLECTION);
            put(ONE_TO_N,TypeFactory.COLLECTION);
        }};
        
        
    // -------------------------------------------------------------------
    // Implementation specific
    // -------------------------------------------------------------------
    
    /**
     * Traverse the component hierarchy and generate the factory for each
     * component contained in the hierarchy.
     */
    private void generateIOE(
        org.objectweb.fractal.juliac.desc.ComponentDesc cdesc )
    throws IOException {
        
        Type type = cdesc.getCT();
        Object controllerDesc = cdesc.getCtrlDesc();
        Object contentDesc = cdesc.getContentClassName();
        FCSourceCodeGeneratorItf<?> fcscg =
            jc.getFCSourceCodeGenerator(type,controllerDesc,contentDesc);
        fcscg.generate(type,controllerDesc,contentDesc);
        
        /*
         * Recurse for subcomponents.
         */
        List<org.objectweb.fractal.juliac.desc.ComponentDesc> subs =
            cdesc.getSubComponents();
        for (org.objectweb.fractal.juliac.desc.ComponentDesc sub : subs) {
            generateIOE(sub);
        }
    }
}
